@@ -28,21 +28,18 @@ import (
 	"k8s.io/autoscaler/cluster-autoscaler/context"
 	"k8s.io/autoscaler/cluster-autoscaler/simulator"
 	"k8s.io/autoscaler/cluster-autoscaler/utils/deletetaint"
-	scheduler_util "k8s.io/autoscaler/cluster-autoscaler/utils/scheduler"
+	kube_util "k8s.io/autoscaler/cluster-autoscaler/utils/kubernetes"
 	. "k8s.io/autoscaler/cluster-autoscaler/utils/test"
 
+	appsv1 "k8s.io/api/apps/v1"
 	apiv1 "k8s.io/api/core/v1"
-	extensionsv1 "k8s.io/api/extensions/v1beta1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
-	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/client-go/kubernetes/fake"
-	core "k8s.io/client-go/testing"
 	kube_record "k8s.io/client-go/tools/record"
 	"k8s.io/kubernetes/pkg/api/testapi"
-	kubeletapis "k8s.io/kubernetes/pkg/kubelet/apis"
 
 	"github.com/stretchr/testify/assert"
-	schedulercache "k8s.io/kubernetes/pkg/scheduler/cache"
+	schedulernodeinfo "k8s.io/kubernetes/pkg/scheduler/nodeinfo"
 )
 
 const MiB = 1024 * 1024
@@ -124,7 +121,88 @@ func TestPodSchedulableMap(t *testing.T) {
 	assert.Nil(t, err)
 }
 
-func TestFilterOutSchedulable(t *testing.T) {
+func TestFilterOutSchedulableByPacking(t *testing.T) {
+	rc1 := apiv1.ReplicationController{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      "rc1",
+			Namespace: "default",
+			SelfLink:  testapi.Default.SelfLink("replicationcontrollers", "rc"),
+			UID:       "12345678-1234-1234-1234-123456789012",
+		},
+	}
+
+	rc2 := apiv1.ReplicationController{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      "rc2",
+			Namespace: "default",
+			SelfLink:  testapi.Default.SelfLink("replicationcontrollers", "rc"),
+			UID:       "12345678-1234-1234-1234-12345678901a",
+		},
+	}
+
+	p1 := BuildTestPod("p1", 1500, 200000)
+	p2_1 := BuildTestPod("p2_2", 3000, 200000)
+	p2_1.OwnerReferences = GenerateOwnerReferences(rc1.Name, "ReplicationController", "extensions/v1beta1", rc1.UID)
+	p2_2 := BuildTestPod("p2_2", 3000, 200000)
+	p2_2.OwnerReferences = GenerateOwnerReferences(rc1.Name, "ReplicationController", "extensions/v1beta1", rc1.UID)
+	p3_1 := BuildTestPod("p3", 300, 200000)
+	p3_1.OwnerReferences = GenerateOwnerReferences(rc2.Name, "ReplicationController", "extensions/v1beta1", rc2.UID)
+	p3_2 := BuildTestPod("p3", 300, 200000)
+	p3_2.OwnerReferences = GenerateOwnerReferences(rc2.Name, "ReplicationController", "extensions/v1beta1", rc2.UID)
+	unschedulablePods := []*apiv1.Pod{p1, p2_1, p2_2, p3_1, p3_2}
+
+	scheduledPod1 := BuildTestPod("s1", 100, 200000)
+	scheduledPod2 := BuildTestPod("s2", 1500, 200000)
+	scheduledPod3 := BuildTestPod("s3", 4000, 200000)
+	var priority1 int32 = 1
+	scheduledPod3.Spec.Priority = &priority1
+	scheduledPod1.Spec.NodeName = "node1"
+	scheduledPod2.Spec.NodeName = "node1"
+	scheduledPod2.Spec.NodeName = "node1"
+
+	podWaitingForPreemption := BuildTestPod("w1", 1500, 200000)
+	var priority100 int32 = 100
+	podWaitingForPreemption.Spec.Priority = &priority100
+	podWaitingForPreemption.Status.NominatedNodeName = "node1"
+
+	p4 := BuildTestPod("p4", 1800, 200000)
+	p4.Spec.Priority = &priority100
+
+	node := BuildTestNode("node1", 2000, 2000000)
+	SetNodeReadyState(node, true, time.Time{})
+
+	predicateChecker := simulator.NewTestPredicateChecker()
+
+	res := filterOutSchedulableByPacking(unschedulablePods, []*apiv1.Node{node}, []*apiv1.Pod{scheduledPod1, scheduledPod3}, []*apiv1.Pod{}, predicateChecker, 10)
+	assert.Equal(t, 3, len(res))
+	assert.Equal(t, p2_1, res[0])
+	assert.Equal(t, p2_2, res[1])
+	assert.Equal(t, p3_2, res[2])
+
+	res2 := filterOutSchedulableByPacking(unschedulablePods, []*apiv1.Node{node}, []*apiv1.Pod{scheduledPod1, scheduledPod2, scheduledPod3}, []*apiv1.Pod{}, predicateChecker, 10)
+	assert.Equal(t, 4, len(res2))
+	assert.Equal(t, p1, res2[0])
+	assert.Equal(t, p2_1, res2[1])
+	assert.Equal(t, p2_2, res2[2])
+	assert.Equal(t, p3_2, res2[3])
+
+	res3 := filterOutSchedulableByPacking(unschedulablePods, []*apiv1.Node{node}, []*apiv1.Pod{scheduledPod1, scheduledPod3}, []*apiv1.Pod{podWaitingForPreemption}, predicateChecker, 10)
+	assert.Equal(t, 4, len(res3))
+	assert.Equal(t, p1, res3[0])
+	assert.Equal(t, p2_1, res3[1])
+	assert.Equal(t, p2_2, res3[2])
+	assert.Equal(t, p3_2, res3[3])
+
+	res4 := filterOutSchedulableByPacking(append(unschedulablePods, p4), []*apiv1.Node{node}, []*apiv1.Pod{scheduledPod1, scheduledPod3}, []*apiv1.Pod{}, predicateChecker, 10)
+	assert.Equal(t, 5, len(res4))
+	assert.Equal(t, p1, res4[0])
+	assert.Equal(t, p2_1, res4[1])
+	assert.Equal(t, p2_2, res4[2])
+	assert.Equal(t, p3_1, res4[3])
+	assert.Equal(t, p3_2, res4[4])
+}
+
+func TestFilterOutSchedulableSimple(t *testing.T) {
 	rc1 := apiv1.ReplicationController{
 		ObjectMeta: metav1.ObjectMeta{
 			Name:      "rc1",
@@ -166,25 +244,25 @@ func TestFilterOutSchedulable(t *testing.T) {
 	podWaitingForPreemption := BuildTestPod("w1", 1500, 200000)
 	var priority100 int32 = 100
 	podWaitingForPreemption.Spec.Priority = &priority100
-	podWaitingForPreemption.Annotations = map[string]string{scheduler_util.NominatedNodeAnnotationKey: "node1"}
+	podWaitingForPreemption.Status.NominatedNodeName = "node1"
 
 	node := BuildTestNode("node1", 2000, 2000000)
 	SetNodeReadyState(node, true, time.Time{})
 
 	predicateChecker := simulator.NewTestPredicateChecker()
 
-	res := FilterOutSchedulable(unschedulablePods, []*apiv1.Node{node}, []*apiv1.Pod{scheduledPod1, scheduledPod3}, []*apiv1.Pod{}, predicateChecker, 10)
+	res := filterOutSchedulableSimple(unschedulablePods, []*apiv1.Node{node}, []*apiv1.Pod{scheduledPod1, scheduledPod3}, []*apiv1.Pod{}, predicateChecker, 10)
 	assert.Equal(t, 2, len(res))
 	assert.Equal(t, p2_1, res[0])
 	assert.Equal(t, p2_2, res[1])
 
-	res2 := FilterOutSchedulable(unschedulablePods, []*apiv1.Node{node}, []*apiv1.Pod{scheduledPod1, scheduledPod2, scheduledPod3}, []*apiv1.Pod{}, predicateChecker, 10)
+	res2 := filterOutSchedulableSimple(unschedulablePods, []*apiv1.Node{node}, []*apiv1.Pod{scheduledPod1, scheduledPod2, scheduledPod3}, []*apiv1.Pod{}, predicateChecker, 10)
 	assert.Equal(t, 3, len(res2))
 	assert.Equal(t, p1, res2[0])
 	assert.Equal(t, p2_1, res2[1])
 	assert.Equal(t, p2_2, res2[2])
 
-	res3 := FilterOutSchedulable(unschedulablePods, []*apiv1.Node{node}, []*apiv1.Pod{scheduledPod1, scheduledPod3}, []*apiv1.Pod{podWaitingForPreemption}, predicateChecker, 10)
+	res3 := filterOutSchedulableSimple(unschedulablePods, []*apiv1.Node{node}, []*apiv1.Pod{scheduledPod1, scheduledPod3}, []*apiv1.Pod{podWaitingForPreemption}, predicateChecker, 10)
 	assert.Equal(t, 3, len(res3))
 	assert.Equal(t, p1, res3[0])
 	assert.Equal(t, p2_1, res3[1])
@@ -202,12 +280,12 @@ func TestFilterOutExpendableAndSplit(t *testing.T) {
 
 	podWaitingForPreemption1 := BuildTestPod("w1", 1000, 200000)
 	podWaitingForPreemption1.Spec.Priority = &priority1
-	podWaitingForPreemption1.Annotations = map[string]string{scheduler_util.NominatedNodeAnnotationKey: "node1"}
+	podWaitingForPreemption1.Status.NominatedNodeName = "node1"
 	podWaitingForPreemption2 := BuildTestPod("w2", 1000, 200000)
 	podWaitingForPreemption2.Spec.Priority = &priority100
-	podWaitingForPreemption2.Annotations = map[string]string{scheduler_util.NominatedNodeAnnotationKey: "node1"}
+	podWaitingForPreemption2.Status.NominatedNodeName = "node1"
 
-	res1, res2 := FilterOutExpendableAndSplit([]*apiv1.Pod{p1, p2, podWaitingForPreemption1, podWaitingForPreemption2}, 0)
+	res1, res2 := filterOutExpendableAndSplit([]*apiv1.Pod{p1, p2, podWaitingForPreemption1, podWaitingForPreemption2}, 0)
 	assert.Equal(t, 2, len(res1))
 	assert.Equal(t, p1, res1[0])
 	assert.Equal(t, p2, res1[1])
@@ -215,7 +293,7 @@ func TestFilterOutExpendableAndSplit(t *testing.T) {
 	assert.Equal(t, podWaitingForPreemption1, res2[0])
 	assert.Equal(t, podWaitingForPreemption2, res2[1])
 
-	res1, res2 = FilterOutExpendableAndSplit([]*apiv1.Pod{p1, p2, podWaitingForPreemption1, podWaitingForPreemption2}, 10)
+	res1, res2 = filterOutExpendableAndSplit([]*apiv1.Pod{p1, p2, podWaitingForPreemption1, podWaitingForPreemption2}, 10)
 	assert.Equal(t, 1, len(res1))
 	assert.Equal(t, p2, res1[0])
 	assert.Equal(t, 1, len(res2))
@@ -229,14 +307,14 @@ func TestFilterOutExpendablePods(t *testing.T) {
 	podWaitingForPreemption1 := BuildTestPod("w1", 1500, 200000)
 	var priority1 int32 = -10
 	podWaitingForPreemption1.Spec.Priority = &priority1
-	podWaitingForPreemption1.Annotations = map[string]string{scheduler_util.NominatedNodeAnnotationKey: "node1"}
+	podWaitingForPreemption1.Status.NominatedNodeName = "node1"
 
 	podWaitingForPreemption2 := BuildTestPod("w1", 1500, 200000)
 	var priority2 int32 = 10
 	podWaitingForPreemption2.Spec.Priority = &priority2
-	podWaitingForPreemption2.Annotations = map[string]string{scheduler_util.NominatedNodeAnnotationKey: "node1"}
+	podWaitingForPreemption2.Status.NominatedNodeName = "node1"
 
-	res := FilterOutExpendablePods([]*apiv1.Pod{p1, p2, podWaitingForPreemption1, podWaitingForPreemption2}, 0)
+	res := filterOutExpendablePods([]*apiv1.Pod{p1, p2, podWaitingForPreemption1, podWaitingForPreemption2}, 0)
 	assert.Equal(t, 3, len(res))
 	assert.Equal(t, p1, res[0])
 	assert.Equal(t, p2, res[1])
@@ -275,14 +353,14 @@ func TestFilterSchedulablePodsForNode(t *testing.T) {
 
 	tn := BuildTestNode("T1-abc", 2000, 2000000)
 	SetNodeReadyState(tn, true, time.Time{})
-	tni := schedulercache.NewNodeInfo()
+	tni := schedulernodeinfo.NewNodeInfo()
 	tni.SetNode(tn)
 
 	context := &context.AutoscalingContext{
 		PredicateChecker: simulator.NewTestPredicateChecker(),
 	}
 
-	res := CheckPodsSchedulableOnNode(context, unschedulablePods, "T1-abc", tni)
+	res := checkPodsSchedulableOnNode(context, unschedulablePods, "T1-abc", tni)
 	wantedSchedulable := []*apiv1.Pod{p1, p3_1, p3_2}
 	wantedUnschedulable := []*apiv1.Pod{p2_1, p2_2}
 
@@ -310,13 +388,13 @@ func TestGetNodeInfosForGroups(t *testing.T) {
 	SetNodeReadyState(unready4, false, time.Now())
 
 	tn := BuildTestNode("tn", 5000, 5000)
-	tni := schedulercache.NewNodeInfo()
+	tni := schedulernodeinfo.NewNodeInfo()
 	tni.SetNode(tn)
 
 	// Cloud provider with TemplateNodeInfo implemented.
 	provider1 := testprovider.NewTestAutoprovisioningCloudProvider(
 		nil, nil, nil, nil, nil,
-		map[string]*schedulercache.NodeInfo{"ng3": tni, "ng4": tni})
+		map[string]*schedulernodeinfo.NodeInfo{"ng3": tni, "ng4": tni})
 	provider1.AddNodeGroup("ng1", 1, 10, 1) // Nodegroup with ready node.
 	provider1.AddNode("ng1", ready1)
 	provider1.AddNodeGroup("ng2", 1, 10, 1) // Nodegroup with ready and unready node.
@@ -330,15 +408,13 @@ func TestGetNodeInfosForGroups(t *testing.T) {
 	provider2 := testprovider.NewTestAutoprovisioningCloudProvider(nil, nil, nil, nil, nil, nil)
 	provider2.AddNodeGroup("ng5", 1, 10, 1) // Nodegroup without nodes.
 
-	fakeClient := &fake.Clientset{}
-	fakeClient.Fake.AddReactor("list", "pods", func(action core.Action) (bool, runtime.Object, error) {
-		return true, &apiv1.PodList{Items: []apiv1.Pod{}}, nil
-	})
+	podLister := kube_util.NewTestPodLister([]*apiv1.Pod{})
+	registry := kube_util.NewListerRegistry(nil, nil, podLister, nil, nil, nil, nil, nil, nil, nil)
 
 	predicateChecker := simulator.NewTestPredicateChecker()
 
-	res, err := GetNodeInfosForGroups([]*apiv1.Node{unready4, unready3, ready2, ready1}, nil,
-		provider1, fakeClient, []*extensionsv1.DaemonSet{}, predicateChecker)
+	res, err := getNodeInfosForGroups([]*apiv1.Node{unready4, unready3, ready2, ready1}, nil,
+		provider1, registry, []*appsv1.DaemonSet{}, predicateChecker)
 	assert.NoError(t, err)
 	assert.Equal(t, 4, len(res))
 	info, found := res["ng1"]
@@ -355,8 +431,8 @@ func TestGetNodeInfosForGroups(t *testing.T) {
 	assertEqualNodeCapacities(t, tn, info.Node())
 
 	// Test for a nodegroup without nodes and TemplateNodeInfo not implemented by cloud proivder
-	res, err = GetNodeInfosForGroups([]*apiv1.Node{}, nil, provider2, fakeClient,
-		[]*extensionsv1.DaemonSet{}, predicateChecker)
+	res, err = getNodeInfosForGroups([]*apiv1.Node{}, nil, provider2, registry,
+		[]*appsv1.DaemonSet{}, predicateChecker)
 	assert.NoError(t, err)
 	assert.Equal(t, 0, len(res))
 }
@@ -376,7 +452,7 @@ func TestGetNodeInfosForGroupsCache(t *testing.T) {
 	SetNodeReadyState(ready6, true, time.Now())
 
 	tn := BuildTestNode("tn", 10000, 10000)
-	tni := schedulercache.NewNodeInfo()
+	tni := schedulernodeinfo.NewNodeInfo()
 	tni.SetNode(tn)
 
 	lastDeletedGroup := ""
@@ -388,7 +464,7 @@ func TestGetNodeInfosForGroupsCache(t *testing.T) {
 	// Cloud provider with TemplateNodeInfo implemented.
 	provider1 := testprovider.NewTestAutoprovisioningCloudProvider(
 		nil, nil, nil, onDeleteGroup, nil,
-		map[string]*schedulercache.NodeInfo{"ng3": tni, "ng4": tni})
+		map[string]*schedulernodeinfo.NodeInfo{"ng3": tni, "ng4": tni})
 	provider1.AddNodeGroup("ng1", 1, 10, 1) // Nodegroup with ready node.
 	provider1.AddNode("ng1", ready1)
 	provider1.AddNodeGroup("ng2", 1, 10, 1) // Nodegroup with ready and unready node.
@@ -400,18 +476,16 @@ func TestGetNodeInfosForGroupsCache(t *testing.T) {
 	provider1.AddNodeGroup("ng4", 0, 1000, 0) // Nodegroup without nodes (and 1 previously ready node).
 	provider1.AddNode("ng4", ready6)
 
-	fakeClient := &fake.Clientset{}
-	fakeClient.Fake.AddReactor("list", "pods", func(action core.Action) (bool, runtime.Object, error) {
-		return true, &apiv1.PodList{Items: []apiv1.Pod{}}, nil
-	})
+	podLister := kube_util.NewTestPodLister([]*apiv1.Pod{})
+	registry := kube_util.NewListerRegistry(nil, nil, podLister, nil, nil, nil, nil, nil, nil, nil)
 
 	predicateChecker := simulator.NewTestPredicateChecker()
 
-	nodeInfoCache := make(map[string]*schedulercache.NodeInfo)
+	nodeInfoCache := make(map[string]*schedulernodeinfo.NodeInfo)
 
 	// Fill cache
-	res, err := GetNodeInfosForGroups([]*apiv1.Node{unready4, unready3, ready2, ready1}, nodeInfoCache,
-		provider1, fakeClient, []*extensionsv1.DaemonSet{}, predicateChecker)
+	res, err := getNodeInfosForGroups([]*apiv1.Node{unready4, unready3, ready2, ready1}, nodeInfoCache,
+		provider1, registry, []*appsv1.DaemonSet{}, predicateChecker)
 	assert.NoError(t, err)
 	// Check results
 	assert.Equal(t, 4, len(res))
@@ -445,8 +519,8 @@ func TestGetNodeInfosForGroupsCache(t *testing.T) {
 	assert.Equal(t, "ng3", lastDeletedGroup)
 
 	// Check cache with all nodes removed
-	res, err = GetNodeInfosForGroups([]*apiv1.Node{}, nodeInfoCache,
-		provider1, fakeClient, []*extensionsv1.DaemonSet{}, predicateChecker)
+	res, err = getNodeInfosForGroups([]*apiv1.Node{}, nodeInfoCache,
+		provider1, registry, []*appsv1.DaemonSet{}, predicateChecker)
 	assert.NoError(t, err)
 	// Check results
 	assert.Equal(t, 2, len(res))
@@ -465,13 +539,13 @@ func TestGetNodeInfosForGroupsCache(t *testing.T) {
 	assert.False(t, found)
 
 	// Fill cache manually
-	infoNg4Node6 := schedulercache.NewNodeInfo()
+	infoNg4Node6 := schedulernodeinfo.NewNodeInfo()
 	err2 := infoNg4Node6.SetNode(ready6.DeepCopy())
 	assert.NoError(t, err2)
-	nodeInfoCache = map[string]*schedulercache.NodeInfo{"ng4": infoNg4Node6}
+	nodeInfoCache = map[string]*schedulernodeinfo.NodeInfo{"ng4": infoNg4Node6}
 	// Check if cache was used
-	res, err = GetNodeInfosForGroups([]*apiv1.Node{ready1, ready2}, nodeInfoCache,
-		provider1, fakeClient, []*extensionsv1.DaemonSet{}, predicateChecker)
+	res, err = getNodeInfosForGroups([]*apiv1.Node{ready1, ready2}, nodeInfoCache,
+		provider1, registry, []*appsv1.DaemonSet{}, predicateChecker)
 	assert.NoError(t, err)
 	assert.Equal(t, 2, len(res))
 	info, found = res["ng2"]
@@ -542,7 +616,7 @@ func TestSanitizeNodeInfo(t *testing.T) {
 
 	node := BuildTestNode("node", 1000, 1000)
 
-	nodeInfo := schedulercache.NewNodeInfo(pod)
+	nodeInfo := schedulernodeinfo.NewNodeInfo(pod)
 	nodeInfo.SetNode(node)
 
 	res, err := sanitizeNodeInfo(nodeInfo, "test-group")
@@ -553,15 +627,15 @@ func TestSanitizeNodeInfo(t *testing.T) {
 func TestSanitizeLabels(t *testing.T) {
 	oldNode := BuildTestNode("ng1-1", 1000, 1000)
 	oldNode.Labels = map[string]string{
-		kubeletapis.LabelHostname: "abc",
-		"x":                       "y",
+		apiv1.LabelHostname: "abc",
+		"x":                 "y",
 	}
 	node, err := sanitizeTemplateNode(oldNode, "bzium")
 	assert.NoError(t, err)
-	assert.NotEqual(t, node.Labels[kubeletapis.LabelHostname], "abc")
+	assert.NotEqual(t, node.Labels[apiv1.LabelHostname], "abc")
 	assert.Equal(t, node.Labels["x"], "y")
 	assert.NotEqual(t, node.Name, oldNode.Name)
-	assert.Equal(t, node.Labels[kubeletapis.LabelHostname], node.Name)
+	assert.Equal(t, node.Labels[apiv1.LabelHostname], node.Name)
 }
 
 func TestSanitizeTaints(t *testing.T) {
